import math


def get_epsilon_of_fedmd_nfdp(n, k, replacement=True):
    """Return epsilon of FedMD-NFDP

    Args:
        n (int): training set size
        k (int): sampling size
        replacement (bool, optional): sampling w/o replacement. Defaults to True.

    Returns:
        float: epsilon of FedMD-NFDP
    """
    if replacement:
        return k * math.log((n + 1) / n)
    else:
        return math.log((n + 1) / (n + 1 - k))


def get_delta_of_fedmd_nfdp(n, k, replacement=True):
    """Return delta of FedMD-NFDP

    Args:
        n (int): training set size
        k (int): sampling size
        replacement (bool, optional): sampling w/o replacement. Defaults to True.

    Returns:
        float: delta of FedMD-NFDP
    """
    if replacement:
        return 1 - ((n - 1) / n) ** k
    else:
        return k / n


def get_k_of_fedmd_nfdp(epsilon, n, replacement=True):
    """Return k of FedMD-NFDP

    Args:
        epsilon (float): epsilon
        n (int): training set size
        replacement (bool, optional): sampling w/o replacement. Defaults to True.

    Returns:
        int: k
    """
    if replacement:
        return int(epsilon / math.log((n + 1) / n))
    else:
        return int((n + 1) * (1 - 1 / math.exp(epsilon)))



import numpy as np

from ...manager import BaseManager
from .torch_wrapper import PaillierTensor


def attach_paillier_to_client_for_encrypted_grad(cls, pk, sk):
    """Makes the client class communicate the encrypted gradients with paillier encryption scheme.

    Args:
        cls: client class
        pk: public key
        sk: secret key
    """

    class PaillierClientWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(PaillierClientWrapper, self).__init__(*args, **kwargs)

        def upload_gradients(self):
            """Uploads encrypted gradients"""
            pt_grads = super().upload_gradients()
            return [
                PaillierTensor(
                    np.vectorize(lambda x: pk.encrypt(x))(grad.detach().numpy())
                )
                for grad in pt_grads
            ]

        def download(self, global_grad):
            """Downloads and decrypt the received global gradients"""
            if not self.initialized:
                # initial parameters are not encrypted
                return super().download(global_grad)
            else:
                decrypted_global_grad = []
                for grad in global_grad:
                    if type(grad) == PaillierTensor:
                        decrypted_global_grad.append(grad.decrypt(sk, self.device))
                    else:
                        decrypted_global_grad.append(grad)
                return super().download(decrypted_global_grad)

    return PaillierClientWrapper


class PaillierGradientClientManager(BaseManager):
    """Client Manager for secure aggregation with Paillier Encryption"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_paillier_to_client_for_encrypted_grad(
            cls, *self.args, **self.kwargs
        )



from .accountant import (  # noqa: F401
    GeneralMomentAccountant,
    culc_tightupperbound_lowerbound_of_rdp_with_theorem6and8_of_zhu_2019,
    culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism,
)
from .client import DPSGDClientManager  # noqa: F401
from .dp_manager import DPSGDManager  # noqa: F401



import copy

import torch

from ...manager import BaseManager


def attach_freerider_to_client(cls, mu, sigma):
    """Wraps the given class in FreeRiderClientWrapper.

    Args:
        mu (float): mean of the gaussian distribution used to generate fake gradients
        sigma (float): standard deviation of the gaussian distribution used to generate fake gradients

    Returns:
        cls: a class wrapped in FreeRiderClientWrapper
    """

    class FreeRiderClientWrapper(cls):
        """Implementation of Free Rider Attack (https://arxiv.org/abs/1911.12560)"""

        def __init__(self, *args, **kwargs):
            super(FreeRiderClientWrapper, self).__init__(*args, **kwargs)
            self.prev_parameters_to_generate_fake_gradients = None

        def upload_gradients(self):
            """Uploads the fake gradients"""
            gradients = []
            if self.prev_parameters_to_generate_fake_gradients is not None:
                for param, prev_param in zip(
                    self.model.parameters(),
                    self.prev_parameters_to_generate_fake_gradients,
                ):
                    gradients.append(
                        (prev_param - param) / self.lr
                        + (sigma * torch.randn_like(param) + mu)
                    )
            else:
                for param in self.model.parameters():
                    gradients.append(sigma * torch.randn_like(param) + mu)
            return gradients

        def download(self, new_global_model):
            """Downloads the new global model"""
            self.prev_parameters_to_generate_fake_gradients = copy.deepcopy(
                list(self.model.parameters())
            )
            super().download(new_global_model)

    return FreeRiderClientWrapper


class FreeRiderClientManager(BaseManager):
    """Manager class for Free-Rider Attack (https://arxiv.org/abs/1911.12560)"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        """Wraps the given class in FreeRiderClientWrapper.

        Returns:
            cls: a class wrapped in FreeRiderClientWrapper
        """
        return attach_freerider_to_client(cls, *self.args, **self.kwargs)



import torch


def attach_differential_privacy_mechanism(
    cls, accountant, l2_norm_clip, noise_multiplier, lot_size, batch_size, dataset_size
):
    """Wraps the given optimizer class in DPOptimizerWrapper.

    Args:
        accountant (BaseMomentAccountant): moment accountant
        l2_norm_clip (float): upper bound of l2-norm
        noise_multiplier (float): scale for added noise
        lot_size (int): sampled lot size
        batch_size (int): batch size
        dataset_size (int): total number of samples in the dataset

    Raises:
        ValueError: if noise_multiplier < 0.0
        ValueError: if l2_norm_clip < 0

    Returns:
        cls: wrapped DPOptimizerWrapper
    """

    class DPOptimizerWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(DPOptimizerWrapper, self).__init__(*args, **kwargs)

            if noise_multiplier < 0.0:
                raise ValueError(
                    "Invalid noise_multiplier: {}".format(noise_multiplier)
                )
            if l2_norm_clip < 0.0:
                raise ValueError("Invalid l2_norm_clip: {}".format(l2_norm_clip))

            self.accountant = accountant
            self.l2_norm_clip = l2_norm_clip
            self.noise_multiplier = noise_multiplier
            self.batch_size = batch_size
            self.lot_size = lot_size

            for group in self.param_groups:
                group["accum_grads"] = [
                    torch.zeros_like(param.data) if param.requires_grad else None
                    for param in group["params"]
                ]

        def zero_grad_keep_accum_grads(self):
            super(DPOptimizerWrapper, self).zero_grad()

        def zero_grad(self):
            self.zero_grad_keep_accum_grads()

        def update_accum_grads(self):
            total_norm = 0.0
            for group in self.param_groups:
                for param in group["params"]:
                    if param.requires_grad:
                        total_norm += param.grad.data.norm(2).item() ** 2.0
            total_norm = total_norm**0.5
            clip_coef = min(self.l2_norm_clip / (total_norm + 1e-6), 1.0)

            for group in self.param_groups:
                for param, accum_grad in zip(group["params"], group["accum_grads"]):
                    if param.requires_grad:
                        accum_grad.add_(param.grad.data.mul(clip_coef))

        def step(self):
            self.update_accum_grads()

        def zero_grad_for_lot(self):
            for group in self.param_groups:
                for accum_grad in group["accum_grads"]:
                    if accum_grad is not None:
                        accum_grad.zero_()

        def step_for_lot(self, *args, **kwargs):
            for group in self.param_groups:
                for param, accum_grad in zip(group["params"], group["accum_grads"]):
                    if param.requires_grad:
                        param.grad.data = accum_grad.clone()
                        param.grad.data.add_(
                            self.l2_norm_clip
                            * self.noise_multiplier
                            * torch.randn_like(param.grad.data)
                        )
                        param.grad.data.mul_(self.batch_size / self.lot_size)
            super(DPOptimizerWrapper, self).step(*args, **kwargs)
            accountant.add_step_info(
                {"sigma": self.noise_multiplier}, self.lot_size / dataset_size, 1
            )

    return DPOptimizerWrapper



import torch

from ...manager import BaseManager


def attach_sparse_gradient_to_client(cls, k):
    """Make the client class communicate the sparse gradients.

    Args:
        cls: client class
        k (int): strength of sparcity

    Returns:
        cls: a class wrapped in SparseGradientClientWrapper
    """

    class SparseGradientClientWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(SparseGradientClientWrapper, self).__init__(*args, **kwargs)

        def upload_gradients(self):
            """Upload sparse gradients"""
            vanila_gradients = super().upload_gradients()
            sparse_gradients = []
            sparse_indices = []
            for vanila_grad in vanila_gradients:
                temp_grad = vanila_grad.reshape(-1)
                # only send top-k gradients
                topk_indices = torch.topk(
                    torch.abs(temp_grad), k=int(len(temp_grad) * k)
                ).indices
                sparse_gradients.append(temp_grad[topk_indices].tolist())
                sparse_indices.append(topk_indices.tolist())

            return sparse_gradients, sparse_indices

    return SparseGradientClientWrapper


def attach_sparse_gradient_to_server(cls):
    """Make the server class communicate the sparse gradients.

    Args:
        cls: server class

    Returns:
        cls: a class wrapped in SparseGradientServerWrapper
    """

    class SparseGradientServerWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(SparseGradientServerWrapper, self).__init__(*args, **kwargs)

        def _preprocess_local_gradients(self, uploaded_grad):
            """Reconstructs dense gradient from the received sparse gradients

            Args:
                uploaded_grad (tuple(torch.Tensor, torch.Tensor)): tuple of non-zero gradients and their positions

            Returns:
                List[torch.Tensor]: list of recovered dense gradients
            """
            sparse_gradients_flattend, sparse_indices = uploaded_grad
            gradients_reshaped = []
            for params, grad, idx in zip(
                self.server_model.parameters(),
                sparse_gradients_flattend,
                sparse_indices,
            ):
                temp_grad = torch.zeros_like(params).reshape(-1)
                temp_grad[idx] = torch.Tensor(grad).to(self.device)
                gradients_reshaped.append(temp_grad.reshape(params.shape))

            return gradients_reshaped

    return SparseGradientServerWrapper


class SparseGradientClientManager(BaseManager):
    """Client-side Manager for sparse gradients."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_sparse_gradient_to_client(cls, *self.args, **self.kwargs)


class SparseGradientServerManager(BaseManager):
    """Server-side Manager for sparse gradients."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_sparse_gradient_to_server(cls, *self.args, **self.kwargs)



import torch


def KL_between_normals(mu_q, sigma_q, mu_p, sigma_p):
    k = mu_q.size(1)

    mu_diff = mu_p - mu_q
    mu_diff_sq = torch.mul(mu_diff, mu_diff)
    logdet_sigma_q = torch.sum(2 * torch.log(torch.clamp(sigma_q, min=1e-8)), dim=1)
    logdet_sigma_p = torch.sum(2 * torch.log(torch.clamp(sigma_p, min=1e-8)), dim=1)

    fs = torch.sum(torch.div(sigma_q**2, sigma_p**2), dim=1) + torch.sum(
        torch.div(mu_diff_sq, sigma_p**2), dim=1
    )
    two_kl = fs - k + logdet_sigma_p - logdet_sigma_q
    return two_kl * 0.5


def mib_loss(
    y,
    sampled_y_pred,
    p_z_given_x_mu,
    p_z_given_x_sigma,
    approximated_z_mean,
    approximated_z_sigma,
    beta=1e-3,
):
    """Implementation of MID loss for NN proposed in https://arxiv.org/abs/2009.05241

    Args:
        y (torch.Tensor): ground-truth label
        sampled_y_pred (torch.Tensor): prdicted output
        p_z_given_x_mu (torch.Tensor): mean of z|x
        p_z_given_x_sigma (torch.Tensor): standard deviation of z|x
        approximated_z_mean (torch.Tensor): approximated mean of z
        approximated_z_sigma (torch.Tensor): approximated standard deviation of z
        beta (float, optional): _description_. weight of I(Z|X) to 1e-3.

    Returns:
        float, torch.Tensor, float: loss values, bound of I(Z|Y) and bound of I(Z|X)
    """

    I_ZX_bound = torch.mean(
        KL_between_normals(
            p_z_given_x_mu, p_z_given_x_sigma, approximated_z_mean, approximated_z_sigma
        )
    )

    loss = torch.nn.CrossEntropyLoss(reduction="none")
    cross_entropy_loss = loss(
        sampled_y_pred, y[:, None].expand(-1, sampled_y_pred.size()[-1])
    )

    cross_entropy_loss_mc = torch.mean(cross_entropy_loss, dim=-1)
    minus_I_ZY_bound = torch.mean(cross_entropy_loss_mc, dim=0)

    return (
        torch.mean(minus_I_ZY_bound + beta * I_ZX_bound),
        minus_I_ZY_bound,
        I_ZX_bound,
    )



from ...manager import BaseManager
from ..core import BaseServer
from ..core.utils import GLOBAL_LOGIT_TAG, LOCAL_LOGIT_TAG


class FedMDServer(BaseServer):
    def __init__(
        self,
        clients,
        server_model=None,
        server_id=0,
        device="cpu",
    ):
        super(FedMDServer, self).__init__(clients, server_model, server_id=server_id)
        self.device = device

        self.uploaded_logits = []

    def forward(self, x):
        if self.server_model is not None:
            return self.server_model(x)
        else:
            return None

    def action(self):
        self.receive()
        self.update()
        self.distribute()

    def receive(self):
        self.uploaded_logits = [client.upload() for client in self.clients]

    def update(self):
        self.consensus = self.uploaded_logits[0]
        len_clients = len(self.clients)
        for logit in self.uploaded_logits[1:]:
            self.consensus += logit / len_clients

    def distribute(self):
        """Distribute the logits of public dataset to each client."""
        for client in self.clients:
            client.download(self.consensus)


def attach_mpi_to_fedmdserver(cls):
    class MPIFedMDServerWrapper(cls):
        """MPI Wrapper for FedMDServer"""

        def __init__(self, comm, *args, **kwargs):
            super(MPIFedMDServerWrapper, self).__init__(*args, **kwargs)
            self.comm = comm
            self.num_clients = len(self.clients)
            self.round = 0

        def action(self):
            self.mpi_receive()
            self.update()
            self.mpi_distribute()
            self.round += 1

        def mpi_receive(self):
            self.mpi_receive_local_logits()

        def mpi_receive_local_logits(self):
            self.uploaded_logits = []

            while len(self.uploaded_logits) < self.num_clients:
                received_logits = self.comm.recv(tag=LOCAL_LOGIT_TAG)
                self.uploaded_logits.append(received_logits)

        def mpi_distribute(self):
            for client_id in self.clients:
                self.comm.send(self.consensus, dest=client_id, tag=GLOBAL_LOGIT_TAG)

    return MPIFedMDServerWrapper


class MPIFedMDServerManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_mpi_to_fedmdserver(cls, *self.args, **self.kwargs)



"""Subpackage for label leakage attack, which infere the private label information of the training dataset.
"""
from .normattack import (  # noqa: F401
    NormAttackSplitNNManager,
    attach_normattack_to_splitnn,
)



from aijack_cpp_core import (  # noqa: F401
    SecureBoostClassifier as SecureBoostClassifierAPI,
)
from aijack_cpp_core import SecureBoostParty as SecureBoostClient  # noqa: F401
from aijack_cpp_core import XGBoostClassifier as XGBoostClassifierAPI  # noqa: F401
from aijack_cpp_core import XGBoostParty as XGBoostClient  # noqa: F401



import math

import numpy as np
from scipy import special
from scipy.special import logsumexp

from .utils import _log_add, _log_erfc, _log_sub


def eps_gaussian(alpha, params):
    if alpha == math.inf:
        return min(
            4 * (np.exp(eps_gaussian(2, params) - 1)),
            2 * np.exp(eps_gaussian(2, params)),
        )
    return alpha / (2 * (params["sigma"] ** 2))


def eps_laplace(alpha, params):
    if alpha <= 1:
        return 1 / params["b"] + np.exp(-1 / params["b"]) - 1
    elif alpha == math.inf:
        return 1 / params["b"]
    return (1 / (alpha - 1)) * logsumexp(
        [
            np.log(alpha / (2 * alpha - 1)) + (alpha - 1) / params["b"],
            np.log((alpha - 1) / (2 * alpha - 1)) + (-alpha) / params["b"],
        ],
        b=[1, 1],
    )


def eps_randresp(alpha, params):
    if params["p"] == 1 or params["p"] == 0:
        return math.inf
    if alpha <= 1:
        return (2 * params["p"] - 1) * np.log(params["p"] / (1 - params["p"]))
    elif alpha == math.inf:
        return np.abs(np.log((1.0 * params["p"] / (1 - params["p"]))))
    return (1 / (alpha - 1)) * logsumexp(
        [
            alpha * np.log(params["p"]) + (1 - alpha) * np.log(1 - params["p"]),
            alpha * np.log(1 - params["p"]) + (1 - alpha) * params["p"],
        ],
        b=[1, 1],
    )


def culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism(
    alpha, params, sampling_rate, _eps
):
    """Compute log(A_alpha) for any positive finite alpha."""
    if float(alpha).is_integer():
        return culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism_int(
            int(alpha), params, sampling_rate
        )
    else:
        return culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism_float(
            alpha, params, sampling_rate
        )


def culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism_int(
    alpha, params, sampling_rate
):
    """Renyi Differential Privacy of the Sampled Gaussian Mechanism
    3.3 Numerically Stable Computatio
    """
    log_a = -np.inf

    for i in range(alpha + 1):
        log_coef_i = (
            np.log(special.binom(alpha, i))
            + i * np.log(sampling_rate)
            + (alpha - i) * np.log(1 - sampling_rate)
        )

        s = log_coef_i + (i * i - i) / (2 * (params["sigma"] ** 2))
        log_a = _log_add(log_a, s)

    return float(log_a) / (alpha - 1)


def culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism_float(
    alpha, params, sampling_rate
):
    """Compute log(A_alpha) for fractional alpha. 0 < q < 1."""
    # The two parts of A_alpha, integrals over (-inf,z0] and [z0, +inf), are
    # initialized to 0 in the log space:
    log_a0, log_a1 = -np.inf, -np.inf
    i = 0

    z0 = params["sigma"] ** 2 * np.log(1 / sampling_rate - 1) + 0.5

    while True:  # do ... until loop
        coef = special.binom(alpha, i)
        log_coef = np.log(abs(coef))
        j = alpha - i

        log_t0 = log_coef + i * np.log(sampling_rate) + j * np.log(1 - sampling_rate)
        log_t1 = log_coef + j * np.log(sampling_rate) + i * np.log(1 - sampling_rate)

        log_e0 = np.log(0.5) + _log_erfc((i - z0) / (math.sqrt(2) * params["sigma"]))
        log_e1 = np.log(0.5) + _log_erfc((z0 - j) / (math.sqrt(2) * params["sigma"]))

        log_s0 = log_t0 + (i * i - i) / (2 * (params["sigma"] ** 2)) + log_e0
        log_s1 = log_t1 + (j * j - j) / (2 * (params["sigma"] ** 2)) + log_e1

        if coef > 0:
            log_a0 = _log_add(log_a0, log_s0)
            log_a1 = _log_add(log_a1, log_s1)
        else:
            log_a0 = _log_sub(log_a0, log_s0)
            log_a1 = _log_sub(log_a1, log_s1)

        i += 1
        if max(log_s0, log_s1) < -30:
            break

    return _log_add(log_a0, log_a1) / (alpha - 1)


def culc_upperbound_of_rdp_with_theorem27_of_wang_2019(
    alpha, params, sampling_rate, _eps
):
    def B(el):
        res = 0
        for i in range(el + 1):
            res += (
                ((-1) ** (i)) * special.binom(el, i) * np.exp((i - 1) * _eps(i, params))
            )
        return abs(res)

    """
    def logAbsB(el):
        ts = []
        ss = []
        for i in range(el + 1):
            ts.append(np.log(special.binom(el, i)) + (i - 1) * _eps(i, params))
            ss.append((-1) ** (i + el % 2))

        return logsumexp(ts, b=ss)
    """

    terms = []
    signs = []

    terms.append(np.log(1))
    signs.append(1)

    second = (
        2 * np.log(sampling_rate)
        + np.log(special.binom(alpha, 2))
        + min(
            np.log(4) + logsumexp([_eps(2, params), np.log(1)], b=[1, -1]),
            _eps(2, params)
            + min(
                np.log(2),
                2 * logsumexp([_eps(math.inf, params), np.log(1)], b=[1, -1]),
            ),
        )
    )
    terms.append(second)
    signs.append(1)

    for j in range(3, alpha + 1):
        terms.append(
            np.log(4)
            + j * np.log(sampling_rate)
            + np.log(special.binom(alpha, j))
            + (1 / 2)
            * (np.log(B(2 * math.floor(j / 2))) + np.log(B(2 * math.ceil(j / 2))))
        )
        signs.append(1)

    return (1 / (alpha - 1)) * logsumexp(terms, b=signs)


def culc_general_upperbound_of_rdp_with_theorem5_of_zhu_2019(
    alpha, params, sampling_rate, _eps
):

    terms = []
    signs = []

    first = ((alpha - 1) * np.log(1 - sampling_rate)) + np.log(
        alpha * sampling_rate - sampling_rate + 1
    )
    terms.append(first)
    signs.append(1)

    if alpha >= 2:
        second = (
            np.log(special.binom(alpha, 2))
            + (2 * np.log(sampling_rate))
            + ((alpha - 2) * np.log(1 - sampling_rate))
            + _eps(2, params)
        )
        terms.append(second)
        signs.append(1)

    if alpha >= 3:
        for el in range(3, alpha + 1):
            third = np.log(3) + (
                np.log(special.binom(alpha, el))
                + (alpha - el) * np.log(1 - sampling_rate)
                + (el * np.log(sampling_rate))
                + ((el - 1) * _eps(el, params))
            )
            terms.append(third)
            signs.append(1)

    return (1 / (alpha - 1)) * logsumexp(terms, b=signs)


def culc_tightupperbound_lowerbound_of_rdp_with_theorem6and8_of_zhu_2019(
    alpha, params, sampling_rate, _eps
):
    terms = []
    signs = []

    first = ((alpha - 1) * np.log(1 - sampling_rate)) + np.log(
        alpha * sampling_rate - sampling_rate + 1
    )
    terms.append(first)
    signs.append(1)

    if alpha >= 2:
        for el in range(2, alpha + 1):
            second = (
                np.log(special.binom(alpha, el))
                + (el * np.log(sampling_rate))
                + ((alpha - el) * np.log(1 - sampling_rate))
                + ((el - 1) * _eps(el, params))
            )
            terms.append(second)
            signs.append(1)

    return (1 / (alpha - 1)) * logsumexp(terms, b=signs)


def culc_tightupperbound_lowerbound_of_rdp_with_theorem6and8_of_zhu_2019_with_tau_estimation(
    alpha, params, sampling_rate, _eps, tau=10
):

    terms = []
    signs = []

    eps_alpha_minus_tau = _eps(alpha - tau, params)

    first = alpha * np.log(1 - sampling_rate) + logsumexp(
        [np.log(1), -eps_alpha_minus_tau], b=[1, -1]
    )
    terms.append(first)
    signs.append(1)

    second = -eps_alpha_minus_tau + alpha * logsumexp(
        [
            np.log(1),
            np.log(sampling_rate),
            np.log(sampling_rate) + eps_alpha_minus_tau,
        ],
        b=[1, -1, 1],
    )
    terms.append(second)
    signs.append(1)

    for el in range(2, tau):
        third = (
            np.log(special.binom(alpha, el))
            + (alpha - el) * np.log(1 - sampling_rate)
            + el * np.log(sampling_rate)
            + logsumexp(
                [(el - 1) * eps_alpha_minus_tau, (el - 1) * _eps(el, params)], b=[1, -1]
            )
        )
        terms.append(third)
        signs.append(-1)

    for el in range(alpha - tau + 1, alpha + 1):
        fourth = (
            np.log(special.binom(alpha, el))
            + (alpha - el) * np.log(1 - sampling_rate)
            + el * np.log(sampling_rate)
            + logsumexp(
                [(el - 1) * _eps(el, params), (el - 1) * eps_alpha_minus_tau],
                b=[1, -1],
            )
        )
        terms.append(fourth)
        signs.append(1)

    return (1 / (alpha - 1)) * logsumexp(terms, b=signs)



import torch
from matplotlib import pyplot as plt

from ..base_attack import BaseAttacker


class MI_FACE(BaseAttacker):
    """Implementation of model inversion attack
    reference: https://dl.acm.org/doi/pdf/10.1145/2810103.2813677

    Attributes:
        target_model: model of the victim
        input_shape: input shapes of taregt model
        auxterm_func (function): the default is constant function
        process_func (function): the default is identity function
    """

    def __init__(
        self,
        target_model,
        input_shape=(1, 1, 64, 64),
        target_label=0,
        lam=0.01,
        num_itr=100,
        auxterm_func=lambda x: 0,
        process_func=lambda x: x,
        apply_softmax=False,
        device="cpu",
        log_interval=1,
        log_show_img=False,
        show_img_func=lambda x: x * 0.5 + 0.5,
    ):
        """Inits MI_FACE
        Args:
            target_model: model of the victim
            input_shape: input shapes of taregt model
            auxterm_func (function): the default is constant function
            process_func (function): the default is identity function
        """
        super().__init__(target_model)
        self.input_shape = input_shape
        self.target_label = target_label
        self.lam = lam
        self.num_itr = num_itr

        self.auxterm_func = auxterm_func
        self.process_func = process_func
        self.device = device
        self.log_interval = log_interval
        self.log_show_img = log_show_img
        self.apply_softmax = apply_softmax
        self.show_img_func = show_img_func

        self.log_image = []

    def attack(
        self,
        init_x=None,
    ):
        """Execute the model inversion attack on the target model.

        Args:
            target_label (int): taregt label
            lam (float): step size
            num_itr (int): number of iteration

        Returns:
            best_img: inversed image with the best score
            log :
        """
        log = []
        if init_x is None:
            x = torch.zeros(self.input_shape, requires_grad=True).to(self.device)
        else:
            init_x = init_x.to(self.device)
            x = init_x

        best_score = float("inf")
        best_img = None

        for i in range(self.num_itr):
            x = x.detach()
            x.requires_grad = True
            pred = self.target_model(x)
            pred = pred.softmax(dim=1) if self.apply_softmax else pred
            target_pred = pred[:, [self.target_label]]
            c = 1 - target_pred + self.auxterm_func(x)
            c.backward()
            grad = x.grad

            if c.item() < best_score:
                best_img = x

            with torch.no_grad():
                x -= self.lam * grad
                x = self.process_func(x)
            log.append(c.item())

            if self.log_interval != 0 and i % self.log_interval == 0:
                print(f"epoch {i}: {c.item()}")
                self._show_img(x)

            self.log_image.append(x.clone())

        self._show_img(x)

        return best_img, log

    def _show_img(self, x):
        if self.log_show_img:
            if self.input_shape[1] == 1:
                plt.imshow(
                    self.show_img_func(x.detach().cpu().numpy()[0][0]),
                    cmap="gray",
                )
                plt.show()
            else:
                plt.imshow(
                    self.show_img_func(x.detach().cpu().numpy()[0].transpose(1, 2, 0))
                )
                plt.show()



"""Subpackage for evasion attack, which creates a malicious data that the target
machine learning model cannot correctly classify.
"""
from .evasion_attack import Evasion_attack_sklearn  # noqa: F401



from abc import abstractmethod


class BaseFLOptimizer:
    """Basic class for optimizers of the server in Federated Learning.

    Args:
        parameters (List[torch.nn.Parameter]): parameters of the model
        lr (float, optional): learning rate. Defaults to 0.01.
        weight_decay (float, optional): coefficient of weight decay. Defaults to 0.0001.
    """

    def __init__(self, parameters, lr=0.01, weight_decay=0.0001):
        self.parameters = list(parameters)
        self.lr = lr
        self.weight_decay = weight_decay
        self.t = 1

    @abstractmethod
    def step(self, grads):
        """Update the parameters with the give gradient

        Args:
            grads (List[torch.Tensor]): list of gradients
        """
        pass



import torch

from ...utils.metrics import crossentropyloss_between_logits
from ..core import BaseServer


class DSFLServer(BaseServer):
    """Server of DS-FL

    Args:
        clients (Llist[torch.nn.Module]): a list of clients.
        global_model (torch.nn.Module): the global model
        public_dataloader (torch.utils.data.DataLoader): a dataloader of the public dataset
        aggregation (str, optional): the type of the aggregation of the logits. Defaults to "ERA".
        distillation_loss_name (str, optional): the type of the loss function fot the distillation loss.
                                                Defaults to "crossentropy".
        era_temperature (float, optional): the temperature of ERA. Defaults to 0.1.
        server_id (int, optional): the id of this server. Defaults to 0.
        device (str, optional): device type. Defaults to "cpu".
    """

    def __init__(
        self,
        clients,
        global_model,
        public_dataloader,
        aggregation="ERA",
        distillation_loss_name="crossentropy",
        era_temperature=0.1,
        server_id=0,
        device="cpu",
    ):
        """Init DSFLServer"""
        super(DSFLServer, self).__init__(clients, global_model, server_id=server_id)
        self.public_dataloader = public_dataloader
        self.aggregation = aggregation
        self.era_temperature = era_temperature
        self.consensus = None
        self.device = device

        self._set_distillation_loss(distillation_loss_name)

    def _set_distillation_loss(self, name):
        """Setup the loss function for distillation.
        `crossentropy`, `L2` or `L1`.

        Args:
            name (str): type of the function

        Raises:
            NotImplementedError: Raises when `name` is not supported.
        """
        if name == "crossentropy":
            self.distillation_loss = crossentropyloss_between_logits
        elif name == "L2":
            self.distillation_loss = torch.nn.MSELoss()
        elif name == "L1":
            self.distillation_loss = torch.nn.L1Loss()
        else:
            raise NotImplementedError(f"{name} is not supported")

    def action(self):
        self.update()
        self.distribute()

    def update(self):
        """Update the aggregated consensus logits with the output logits received from the clients.

        Raises:
            NotImplementedError: Raises when the specified aggregation type is not supported.
        """
        if self.aggregation == "ERA":
            self._entropy_reduction_aggregation()
        elif self.aggregation == "SA":
            self._simple_aggregation()
        else:
            raise NotImplementedError(f"{self.aggregation} is not supported")

    def update_globalmodel(self, global_optimizer):
        """Train the global model with the global consensus logits.

        Args:
            global_optimizer (torch.optim.Optimizer): an optimizer

        Returns:
            float: average loss
        """
        running_loss = 0
        for global_data in self.public_dataloader:
            idx = global_data[0]
            x = global_data[1].to(self.device)
            y_global = self.consensus[idx, :].to(self.device)
            global_optimizer.zero_grad()
            y_pred = self(x)
            loss_consensus = self.distillation_loss(y_pred, y_global)
            loss_consensus.backward()
            global_optimizer.step()
            running_loss += loss_consensus.item()
        running_loss /= len(self.public_dataloader)
        return running_loss

    def distribute(self):
        """Distribute the logits of public dataset to each client."""
        for client in self.clients:
            client.download(self.consensus)

    def _entropy_reduction_aggregation(self):
        """Aggregate the received logits with ERA"""
        self._simple_aggregation()
        self.consensus = torch.softmax(self.consensus / self.era_temperature, dim=1)

    def _simple_aggregation(self):
        """Aggregate the received logits with SA (calculating average)"""
        self.consensus = self.clients[0].upload() / len(self.clients)
        for client in self.clients[1:]:
            self.consensus += client.upload() / len(self.clients)



import copy
from abc import abstractmethod

from ...utils import accuracy_torch_dataloader


class BaseFedAPI:
    """Abstract class for Federated Learning API"""

    @abstractmethod
    def local_train(self):
        pass

    @abstractmethod
    def run(self):
        pass


class BaseFLKnowledgeDistillationAPI:
    """Abstract class for API of federated learning with knowledge distillation.

    Args:
        server (aijack.collaborative.core.BaseServer): the server
        clients (List[aijack.collaborative.core.BaseClient]): a list of the clients
        public_dataloader (torch.utils.data.DataLoader): a dataloader for the public dataset
        local_dataloaders (List[torch.utils.data.DataLoader]): a list of local dataloaders
        validation_dataloader (torch.utils.data.DataLoader): a dataloader for the validation dataset
        criterion (function): a function to calculate the loss
        num_communication (int): the number of communication
        device (str): device type
    """

    def __init__(
        self,
        server,
        clients,
        public_dataloader,
        local_dataloaders,
        validation_dataloader,
        criterion,
        num_communication,
        device,
    ):
        """Initialize BaseFLKnowledgeDistillationAPI"""
        self.server = server
        self.clients = clients
        self.public_dataloader = public_dataloader
        self.local_dataloaders = local_dataloaders
        self.validation_dataloader = validation_dataloader
        self.criterion = criterion
        self.num_communication = num_communication
        self.device = device

        self.client_num = len(clients)

    def train_client(self, epoch=1, public=True):
        """Train local models with the local datasets or the public dataset.

        Args:
            public (bool, optional): Train with the public dataset or the local datasets.
                                     Defaults to True.

        Returns:
            List[float]: a list of average loss of each clients.
        """
        loss_on_local_dataest = []
        for client_idx in range(self.client_num):
            if public:
                trainloader = self.public_dataloader
            else:
                trainloader = self.local_dataloaders[client_idx]

            running_loss = self.clients[client_idx].local_train(
                epoch, self.criterion, trainloader, self.client_optimizers[client_idx]
            )

            loss_on_local_dataest.append(copy.deepcopy(running_loss / len(trainloader)))

        return loss_on_local_dataest

    @abstractmethod
    def run(self):
        pass

    def score(self, dataloader, only_local=False):
        """Returns the performance on the given dataset.

        Args:
            dataloader (torch.utils.data.DataLoader): a dataloader
            only_local (bool): show only the local results

        Returns:
            Dict[str, int]: performance of global model and local models
        """

        clients_score = [
            accuracy_torch_dataloader(client, dataloader, device=self.device)
            for client in self.clients
        ]

        if only_local:
            return {"clients_score": clients_score}
        else:
            server_score = accuracy_torch_dataloader(
                self.server, dataloader, device=self.device
            )
            return {"server_score": server_score, "clients_score": clients_score}

    def local_score(self):
        """Returns the local performance of each clients.

        Returns:
            Dict[str, int]: performance of global model and local models
        """
        local_score_list = []
        for client, local_dataloader in zip(self.clients, self.local_dataloaders):
            temp_score = accuracy_torch_dataloader(
                client, local_dataloader, device=self.device
            )
            local_score_list.append(temp_score)

        return {"clients_score": local_score_list}



import torch

from ..fedavg import FedAVGClient


class FedProxClient(FedAVGClient):
    def local_train(
        self,
        server_parameters,
        local_epoch,
        criterion,
        trainloader,
        optimizer,
        communication_id=0,
    ):
        for i in range(local_epoch):
            running_loss = 0.0
            running_data_num = 0
            for _, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs = inputs.to(self.device)
                inputs.requires_grad = True
                labels = labels.to(self.device)

                optimizer.zero_grad()
                self.zero_grad()

                outputs = self(inputs)
                loss = criterion(outputs, labels)
                loss.backward()

                for local_param, global_param in zip(
                    self.parameters(), server_parameters
                ):
                    loss += (
                        self.mu / 2 * torch.norm(local_param.data - global_param.data)
                    )
                    local_param.grad.data += self.mu * (
                        local_param.data - global_param.data
                    )

                optimizer.step()

                running_loss += loss.item()
                running_data_num += inputs.shape[0]

            print(
                f"communication {communication_id}, epoch {i}: client-{self.user_id+1}",
                running_loss / running_data_num,
            )



import numpy as np
import torch
from sklearn.metrics import accuracy_score
from torch.nn import functional as F


def total_variance(x):
    """Returns the total variance of the given data

    Args:
        x (torch.Tensor): input data

    Returns:
        float: total variance of the given data
    """
    dx = torch.mean(torch.abs(x[:, :, :, :-1] - x[:, :, :, 1:]))
    dy = torch.mean(torch.abs(x[:, :, :-1, :] - x[:, :, 1:, :]))
    return dx + dy


def crossentropyloss_between_logits(y_pred_logit, y_true_labels, reduction="mean"):
    """Cross entropy loss for soft labels
    Based on https://discuss.pytorch.org/t/soft-cross-entropy-loss-tf-has-it-does-pytorch-have-it/69501/2
    Args:
        y_pred_logit (torch.Tensor): predicted logits
        y_true_labels (torch.Tensor): ground-truth soft labels

    Returns:
        torch.Tensor: average cross entropy between y_pred_logit and y_true_labels2
    """
    results = -torch.sum(F.log_softmax(y_pred_logit, dim=1) * y_true_labels, dim=1)

    if reduction == "sum":
        return torch.sum(results)
    elif reduction == "mean":
        return torch.mean(results)
    else:
        raise NotImplementedError(f"`reduction`={reduction} is not supported.")


def accuracy_torch_dataloader(model, dataloader, device="cpu", xpos=1, ypos=2):
    """Calculates the accuracy of the model on the given dataloader

    Args:
        model (torch.nn.Module): model to be evaluated
        dataloader (torch.DataLoader): dataloader to be evaluated
        device (str, optional): device type. Defaults to "cpu".
        xpos (int, optional): the positional index of the input in data. Defaults to 1.
        ypos (int, optional): the positional index of the label in data. Defaults to 2.

    Returns:
        float: accuracy
    """
    in_preds = []
    in_label = []
    with torch.no_grad():
        for data in dataloader:
            inputs = data[xpos]
            labels = data[ypos]
            inputs = inputs.to(device)
            labels = labels.to(device).to(torch.int64)
            outputs = model(inputs)
            in_preds.append(outputs)
            in_label.append(labels)
        in_preds = torch.cat(in_preds)
        in_label = torch.cat(in_label)

    return accuracy_score(
        np.array(torch.argmax(in_preds, axis=1).cpu()), np.array(in_label.cpu())
    )



import torch
from torch import nn

from ..core import BaseServer


class FedGEMSServer(BaseServer):
    def __init__(
        self,
        clients,
        global_model,
        len_public_dataloader,
        output_dim=1,
        self_evaluation_func=None,
        base_loss_func=nn.CrossEntropyLoss(),
        kldiv_loss_func=nn.KLDivLoss(),
        server_id=0,
        lr=0.1,
        epsilon=0.75,
        device="cpu",
    ):
        super(FedGEMSServer, self).__init__(clients, global_model, server_id=server_id)
        self.len_public_dataloader = len_public_dataloader
        self.lr = lr
        self.epsilon = epsilon
        self.self_evaluation_func = self_evaluation_func
        self.base_loss_func = base_loss_func
        self.kldiv_loss_func = kldiv_loss_func
        self.output_dim = output_dim
        self.device = device

        self.global_pool_of_logits = torch.ones((len_public_dataloader, output_dim)).to(
            self.device
        ) * float("inf")
        self.predicted_values = torch.ones((len_public_dataloader, output_dim)).to(
            self.device
        ) * float("inf")

    def action(self):
        self.distribute()

    def update(self, idxs, x):
        """Register the predicted logits to self.predicted_values"""
        self.predicted_values[idxs] = self.server_model(x).detach().to(self.device)

    def distribute(self):
        """Distribute the logits of public dataset to each client."""
        for client in self.clients:
            client.download(self.predicted_values)

    def self_evaluation_on_the_public_dataset(self, idxs, x, y):
        """Execute self evaluation on the public dataset

        Args:
            idxs (torch.Tensor): indexs of x
            x (torch.Tensor): input data
            y (torch.Tensr): labels of x

        Returns:
            the loss
        """
        y_pred = self.server_model(x)
        correct_idx, incorrect_idx = self.self_evaluation_func(y_pred, y)

        loss_s1 = 0
        loss_s2 = 0
        loss_s3 = 0

        # for each sample that the server predicts correctly
        if len(correct_idx) != 0:
            loss_s1 += self.base_loss_func(
                y_pred[correct_idx], y[correct_idx].to(torch.int64)
            )
            self.global_pool_of_logits[idxs[correct_idx]] = y_pred[correct_idx].detach()

        # for each sample that the server predicts wrongly
        s_incorrect_not_star_idx = [
            iid.item()
            for iid in incorrect_idx
            if self.global_pool_of_logits[idxs[iid]][0].item() != float("inf")
        ]
        if len(s_incorrect_not_star_idx) != 0:
            loss_s2 += self.epsilon * self.base_loss_func(
                y_pred[s_incorrect_not_star_idx],
                y[s_incorrect_not_star_idx].to(torch.int64),
            ) + (1 - self.epsilon) * self.kldiv_loss_func(
                self.global_pool_of_logits[idxs[s_incorrect_not_star_idx]]
                .softmax(dim=-1)
                .log(),
                y_pred[s_incorrect_not_star_idx].softmax(dim=-1),
            )

        s_incorrect_star_idx = list(
            set(incorrect_idx.cpu().tolist()) - set(s_incorrect_not_star_idx)
        )
        if len(s_incorrect_star_idx) != 0:
            loss_s3 += self.epsilon * self.base_loss_func(
                y_pred[s_incorrect_star_idx], y[s_incorrect_star_idx]
            ) + (1 - self.epsilon) * self.kldiv_loss_func(
                self._get_knowledge_from_clients(
                    x[s_incorrect_star_idx], y[s_incorrect_star_idx]
                )
                .softmax(dim=-1)
                .log(),
                y_pred[s_incorrect_star_idx].softmax(dim=-1),
            )
        loss = loss_s1 + loss_s2 + loss_s3
        return loss

    def _get_knowledge_from_clients(self, x, y):
        client_weight = torch.zeros(self.num_clients, y.shape[0]).to(self.device)
        client_knowledge = torch.zeros(
            self.num_clients, y.shape[0], self.output_dim
        ).to(self.device)
        for cid, client in enumerate(self.clients):
            y_pred = client.upload(x).to(self.device)
            client_knowledge[cid] = y_pred
            correct_idx, _ = self.self_evaluation_func(y_pred, y)
            if len(correct_idx) != 0:
                ep = torch.zeros((y_pred.shape[0])).to(self.device)
                ep[correct_idx] += -1 * torch.sum(
                    y_pred[correct_idx].softmax(dim=-1)
                    * torch.log(y_pred[correct_idx].softmax(dim=-1)),
                    dim=1,
                )
                client_weight[cid, correct_idx] = 1 / ep[correct_idx]

        client_weight = (
            client_weight.softmax(dim=0)
            .reshape(self.num_clients, y.shape[0], 1)
            .expand(self.num_clients, y.shape[0], self.output_dim)
        )

        ensembled_knowledge = torch.sum(
            client_weight * client_knowledge,
            dim=0,
        )

        return ensembled_knowledge.detach()



"""Subpackage for membership inference attack, which reveals the confidential information
    about whether the target data is in the training dataset or not.
"""
from .membership_inference import (  # noqa: F401
    AttackerModel,
    Membership_Inference,
    ShadowModel,
)



import torch


class DataRepExtractor:
    def __init__(self, net, num_fc_layers=1, m=1, bias=True):
        self.net = net
        self.num_fc_layers = num_fc_layers
        self.m = m
        self.bias = bias

        # dl_dw = torch.autograd.grad(loss, net.parameters(), retain_graph=True)
        # dl_dw = [g.detach() for g in dl_dw]
        # extractor = DataRepExtractor(net, num_fc_layers, m, bias)
        # datarep = extractor.extract_datarep(dl_dw)

    def get_dldw(self, loss):
        dldw = torch.autograd.grad(loss, self.net.parameters(), retain_graph=True)
        dldw = [g.detach() for g in dldw]
        return dldw

    def extract_datarep(self, dldw):
        max_idx = torch.argmax(dldw[-2].norm(2, dim=1))
        rep_1 = dldw[-2][max_idx, :].reshape(1, -1)

        reps = [rep_1]
        for i in range(1, self.num_fc_layers):
            grad_idx = -2 * (i + 1) if self.bias else -2 * i + 1
            rep_i = (
                dldw[grad_idx][torch.topk(reps[-1].norm(2, dim=0), self.m)[1], :]
                .mean(dim=0)
                .reshape(1, -1)
            )
            reps.append(rep_i)

        return reps



"""Subpackge implementing abstract classes for collaborative learning.
"""
from .api import BaseFedAPI, BaseFLKnowledgeDistillationAPI  # noqa : F401
from .client import BaseClient  # noqa: F401
from .server import BaseServer  # noqa: F401



from .base import BaseFLOptimizer


class SGDFLOptimizer(BaseFLOptimizer):
    """Implementation of SGD to update the global model of Federated Learning.

    Args:
        parameters (List[torch.nn.Parameter]): parameters of the model
        lr (float, optional): learning rate. Defaults to 0.01.
        weight_decay (float, optional): coefficient of weight decay. Defaults to 0.0001.
    """

    def __init__(self, parameters, lr=0.01, weight_decay=0.0000):
        super().__init__(parameters, lr=lr, weight_decay=weight_decay)

    def step(self, grads):
        """Update the parameters with the give gradient

        Args:
            grads (List[torch.Tensor]): list of gradients
        """
        for param, grad in zip(self.parameters, grads):
            if self.weight_decay == 0.0:
                param.data -= self.lr * grad
            else:
                param.data -= self.lr * (grad + self.weight_decay * param.data)
        self.t += 1



import random

import torch

from ...manager import BaseManager


def attach_label_flip_attack_to_client(
    cls, victim_label, target_label=None, class_num=None
):
    class LabelFlipAttackClientWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(LabelFlipAttackClientWrapper, self).__init__(*args, **kwargs)

        def local_train(
            self, local_epoch, criterion, trainloader, optimizer, communication_id=0
        ):
            for i in range(local_epoch):
                running_loss = 0.0
                running_data_num = 0
                for _, data in enumerate(trainloader, 0):
                    inputs, labels = data
                    inputs = inputs.to(self.device)

                    if target_label is not None:
                        labels = torch.where(
                            labels == victim_label, target_label, labels
                        )
                    else:
                        labels = torch.where(
                            labels == victim_label, random.randint(0, class_num), labels
                        )

                    labels = labels.to(self.device)

                    optimizer.zero_grad()
                    self.zero_grad()

                    outputs = self(inputs)
                    loss = criterion(outputs, labels)

                    loss.backward()
                    optimizer.step()

                    running_loss += loss.item()
                    running_data_num += inputs.shape[0]

                print(
                    f"communication {communication_id}, epoch {i}: client-{self.user_id+1}",
                    running_loss / running_data_num,
                )

    return LabelFlipAttackClientWrapper


class LabelFlipAttackClientManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_label_flip_attack_to_client(cls, *self.args, **self.kwargs)



import torch
from torch import nn

from .loss import mib_loss


class VIB(nn.Module):
    def __init__(self, encoder, decoder, dim_z=256, num_samples=10, beta=1e-3):
        super(VIB, self).__init__()
        self.dim_z = dim_z
        self.num_samples = num_samples
        self.encoder = encoder
        self.decoder = decoder
        self.beta = beta

    def get_params_of_p_z_given_x(self, x):
        encoder_output = self.encoder(x)
        if encoder_output.shape[1] != self.dim_z * 2:
            raise ValueError("the output dimension of encoder must be 2 * dim_z")
        mu = encoder_output[:, : self.dim_z]
        sigma = torch.nn.functional.softplus(encoder_output[:, self.dim_z :])
        return mu, sigma

    def sampling_from_encoder(self, mu, sigma, batch_size):
        return mu + sigma * torch.normal(
            torch.zeros(self.num_samples, batch_size, self.dim_z),
            torch.ones(self.num_samples, batch_size, self.dim_z),
        )

    def forward(self, x):
        batch_size = x.size()[0]

        # encoder
        p_z_given_x_mu, p_z_given_x_sigma = self.get_params_of_p_z_given_x(x)
        sampled_encoded_features = self.sampling_from_encoder(
            p_z_given_x_mu, p_z_given_x_sigma, batch_size
        )

        # decoder
        sampled_decoded_outputs = self.decoder(sampled_encoded_features)
        outputs = torch.mean(sampled_decoded_outputs, dim=0)

        if self.training:
            return outputs, {
                "sampled_decoded_outputs": sampled_decoded_outputs.permute(1, 2, 0),
                "sampled_encoded_features": sampled_encoded_features,
                "p_z_given_x_mu": p_z_given_x_mu,
                "p_z_given_x_sigma": p_z_given_x_sigma,
            }
        else:
            return outputs

    def loss(self, y, result_dict):
        sampled_y_pred = result_dict["sampled_decoded_outputs"]
        p_z_given_x_mu = result_dict["p_z_given_x_mu"]
        p_z_given_x_sigma = result_dict["p_z_given_x_sigma"]

        approximated_z_mean = torch.zeros_like(p_z_given_x_mu)
        approximated_z_sigma = torch.ones_like(p_z_given_x_sigma)

        loss, _, _ = mib_loss(
            y,
            sampled_y_pred,
            p_z_given_x_mu,
            p_z_given_x_sigma,
            approximated_z_mean,
            approximated_z_sigma,
            beta=self.beta,
        )

        return loss



from .dataloader import prepareFederatedMNISTDataloaders  # noqa: F401
from .metrics import (  # noqa: F401
    accuracy_torch_dataloader,
    crossentropyloss_between_logits,
    total_variance,
)
from .utils import torch_round_x_decimal  # noqa: F401
from .utils import NumpyDataset, try_gpu, worker_init_fn  # noqa: F401



import copy

import numpy as np
import sklearn

from ..base_attack import BaseAttacker


class Evasion_attack_sklearn(BaseAttacker):
    """Creates an adversarial example against sklearn objects
        reference https://arxiv.org/abs/1708.06131

    Args:
        target_model (sklearn): sklearn classifier
        X_minus_1 (numpy.array): datasets that contains
                                only the class you want to misclasssify
        dmax (float): max distance between the adversarial example
                        and initial one
        max_iter (int): maxium number of iterations
        gamma (float): parameter gamma of svm (used for only svm)
        lam (float): trade - off parameter
        t (float): step_size
        h (float): a badwidtch paramter for a KDE
        distance (str): type of distance such as L2 or L1
        kde_type (str): type of kernel density estimator

    Attributes:
        target_model (sklearn): sklearn classifier
        X_minus_1 (numpy.array): datasets that contains
                                only the class you want to misclasssify
        dmax (float): max distance between the adversarial example
                        and initial one
        max_iter (int): maxium number of iterations
        gamma (float): parameter gamma of svm (used for only svm)
        lam (float): trade - off parameter
        t (float): step_size
        h (float): a badwidtch paramter for a KDE
        distance (str): type of distance such as L2 or L1
        kde_type (str): type of kernel density estimator
        n_minus_1 (int): number of rows of X_minus_1
        delta_g (func): deviation of he discriminant function of a
                        surrogate classifier f learnt on D

    Raises:
        ValueError: if given distance is not supported.

    Examples:
        >>>X_minus_1 = X_train[np.where(y_train == "3")]
        >>>attacker = Attack_sklearn(target_model = target_model,
                                    X_minus_1 = X_minus_1,
                                    dmax =  (5000 / 255) * 2.5,
                                    max_iter = 300,
                                    gamma = 1 / (X_train.shape[1] *
                                                np.var(X_train)),
                                    lam = 10, t = 0.5, h = 10)
        >>>xm, log = attacker.attack(x0)
    """

    def __init__(
        self,
        target_model,
        X_minus_1,
        dmax,
        max_iter,
        gamma,
        lam,
        t,
        h,
        distance="L1",
        kde_type="L1",
    ):
        super().__init__(target_model)

        self.X_minus_1 = X_minus_1
        self.dmax = dmax
        self.max_iter = max_iter
        self.gamma = gamma
        self.lam = lam
        self.t = t
        self.h = h
        self.kde_type = kde_type

        self.n_minus_1 = X_minus_1.shape[0]

        self.delta_g = None
        self.distance = None

        _ = self._detect_type_of_classifier()

        if distance == "L1":
            self.distance = lambda x1, x2: np.sum(np.abs(x1 - x2))
        else:
            raise ValueError(f"distance type {distance} is not defined")

    def _detect_type_of_classifier(self):
        """Sets proper attributes based on the type of classifier

        Returns:
            return True (bool) if there is no error

        Raises:
            ValueError : if target classifier is not supported
        """

        target_type = type(self.target_model)

        if target_type == sklearn.svm._classes.SVC:
            params = self.target_model.get_params()
            kernel_type = params["kernel"]
            if kernel_type == "rbf":

                def kernel(xm):
                    return np.exp(
                        -self.gamma * ((xm - self.target_model.support_vectors_) ** 2)
                    )

                def delta_kernel(xm):
                    return (
                        (-2 * self.gamma)
                        * kernel(xm)
                        * (xm - self.target_model.support_vectors_)
                    )

            elif kernel_type == "linear":

                def delta_kernel(xm):
                    return self.target_model.support_vectors_

            elif kernel_type == "poly":
                p = params["degree"]
                c = self.target_model.intercept_

                def delta_kernel(xm):
                    return (
                        p
                        * (((xm * self.target_model.support_vectors_) + c) ** (p - 1))
                        * self.target_model.support_vectors_
                    )

            else:
                raise ValueError(f"kernel type {kernel_type} is not supported")

            self.delta_g = lambda xm: self.target_model.dual_coef_.dot(delta_kernel(xm))

        else:
            raise ValueError(f"target type {target_type} is not supported")

        return True

    def _get_delta_p(self, xm):
        """Culculates deviation of the estimated density p(xm−1 |yc = −1)

        Args:
            xm (np.array) : an adversarial example

        Returns:
            delta_p (np.array) : deviation of p

        """

        if self.kde_type == "L1":
            a = -1 / (self.n_minus_1 * self.h)
            b = np.exp(-(np.sum(np.abs(xm - self.X_minus_1), axis=1)) / self.h).dot(
                xm - self.X_minus_1
            )
            delta_p = a * b
            return delta_p

    def _get_grad_f(self, xm, norm="l1"):
        """Culculates deviation of objective function F

        Args:
            xm (np.array) : an adversarial example
            norm (str) : type of distance for normalization

        Returns:
            delta_f (np.array) : deviation of F

        Raises:
            ValueError : if the type of norm is not supported
        """

        delta_f = self.delta_g(xm) - self.lam * self._get_delta_p(xm)
        if norm == "l1":
            delta_f /= np.sum(np.abs(delta_f)) + 1e-5
        elif norm == "l2":
            delta_f /= np.sqrt(np.sum(delta_f**2, axis=0)) + 1e-5
        else:
            raise ValueError(f"norm type {norm} is not defined")

        return delta_f

    def attack(self, x0):
        """Executes evasion attack

        Args:
            x0 (np.array) : initial data point

        Returns:
            xm (np.array) : created adversarial example
            g_list (list) : lof of decision function (only for svm)
                            (need future improvement)
        """

        g_list = []
        xm = copy.copy(x0)
        for i in range(self.max_iter):
            delta_f = self._get_grad_f(xm)
            xm -= self.t * delta_f.reshape(-1)
            d = self.distance(xm, x0)  # + i * (10/255)
            if d > self.dmax:
                xm = x0 + ((xm - x0) / d) * self.dmax

            g_list.append(self.target_model.decision_function(xm.reshape(1, -1)))

        return xm, g_list



from ...manager import BaseManager


def attach_history_attack_to_client(cls, lam):
    class HistoryAttackClientWrapper(cls):
        """Implementation of history attack proposed in https://arxiv.org/pdf/2203.08669.pdf"""

        def __init__(self, *args, **kwargs):
            super(HistoryAttackClientWrapper, self).__init__(*args, **kwargs)

        def upload_gradients(self):
            """Upload the local gradients"""
            gradients = []
            for param, prev_param in zip(self.model.parameters(), self.prev_parameters):
                gradients.append((param - prev_param) * lam)
            return gradients

    return HistoryAttackClientWrapper


class HistoryAttackClientWrapper(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_history_attack_to_client(cls, *self.args, **self.kwargs)



"""Subpackage for model inversion attack, which reconstructs the private data from
the trained machine learning models.
"""
from .gan_attack import GANAttackClientManager, attach_ganattack_to_client  # noqa: F401
from .generator_attack import Generator_Attack  # noqa: F401
from .gradientinversion import (  # noqa: F401
    GradientInversion_Attack,
    GradientInversionAttackServerManager,
    attach_gradient_inversion_attack_to_server,
)
from .mi_face import MI_FACE  # noqa: F401
from .utils import DataRepExtractor  # noqa: F401



"""Subpackage for defense algorithms for machine learning models.
"""
from .ckks import CKKSEncoder, CKKSEncrypter, CKKSPlaintext  # noqa: F401
from .dp import DPSGDManager, GeneralMomentAccountant  # noqa: F401
from .mid import VIB, KL_between_normals, mib_loss  # noqa:F401
from .paillier import PaillierGradientClientManager, PaillierKeyGenerator  # noqa: F401
from .purifier import Purifier_Cifar10, PurifierLoss  # noqa: F401
from .soteria import SoteriaClientManager, attach_soteria_to_client  # noqa: F401
